#!/usr/bin/env python3
# encoding utf-8

#coding=utf-8

# from DiscreteHFO.HFOAttackingPlayer import HFOAttackingPlayer
# from DiscreteHFO.Agent import Agent
import argparse
from collections import defaultdict
import numpy as np
from hfoEnv import hfoEnv


class MonteCarloAgent(object):
    def __init__(self, discountFactor, epsilon, initVals=0.0):
        super(MonteCarloAgent, self).__init__()
        self.discountFactor = discountFactor
        self.epsilon = epsilon
        self.currentState = None
        self.actions = [
            "DRIBBLE_UP",
            "DRIBBLE_DOWN",
            "DRIBBLE_LEFT",
            "DRIBBLE_RIGHT",
            "SHOOT",
        ]
        self.Qs = defaultdict(lambda: np.ones(len(self.actions)) * initVals)
        self.random = np.random.RandomState(0)
        self.policy             = defaultdict(lambda: np.ones((len(self.actions))) / len(self.actions))
        self.state_action_count = defaultdict(lambda: np.zeros((len(self.actions))))

    def learn(self):
        state_action_count = defaultdict(lambda: np.zeros((len(self.actions))))  # default value: [0, 0, 0, 0, 0]
        states = self.cache["states"]
        actions = self.cache["actions"]
        rewards = self.cache["rewards"]
        statuses = self.cache["statuses"]
        nextStates = self.cache["nextStates"]
        results = []

        for idx, (state, action, _, _, _) in enumerate (
            zip(states, actions, rewards, statuses, nextStates)
        ):
            action_idx = self.actions.index(action)
            if state_action_count[state][action_idx] != 0:  # use dict with in key word is better?
                continue

            state_action_count[state][action_idx] += 1
            self.state_action_count[state][action_idx] += 1
            rewards_after_state_action = rewards[idx:] # rewards after state x is rewards of all succeeding states.
            discounted_rewards = sum (  # this solution is n^2, we can improve it
                [self.discountFactor ** idx * reward for idx, reward in enumerate(rewards_after_state_action)]
            )
            old_Q = self.Qs[state][action_idx]
            self.Qs[state][action_idx] = old_Q + (discounted_rewards - old_Q) * (  # update mu incrementally
                1 / self.state_action_count[state][action_idx]
            )
            results.append(self.Qs[state][action_idx])
        for state, Q in self.Qs.items():
            action = np.argmax(Q)
            for a in range(len(self.actions)):
                if a == action:
                    self.policy[state][a] = (
                        1 - self.epsilon + self.epsilon / len(self.actions)
                    )
                else:
                    self.policy[state][a] = self.epsilon / len(self.actions)
        return _, results

    def toStateRepresentation(self, state):
        return str(state)

    def setExperience(self, state, action, reward, status, nextState):
        self.cache["states"].append(state)
        self.cache["actions"].append(action)
        self.cache["rewards"].append(reward)
        self.cache["statuses"].append(status)
        self.cache["nextStates"].append(nextState)


    def setState(self, state):
        self.currentState = state

    def reset(self):
        self.cache = {
            "states": [],
            "actions": [],
            "rewards": [],
            "statuses": [],
            "nextStates": []
        }

    def act(self):  # play the next move using current policy
        idx = np.random.choice(len(self.actions), size=1, p=self.policy[self.currentState])[0]
        return self.actions[idx]

    def setEpsilon(self, epsilon):
        self.epsilon = epsilon

    def computeHyperparameters(self, numTakenActions, episodeNumber):  # explore before iteration 4000. exploit after 4
        """
        compute real epsilon.
        After 4000 iteration, the probability of choice become determined.
        Otherwise, returns epsilon which slightly decreases each turn.
        —— explore before iteration 4000. exploit after 4000
        """
        return 0.0 if episodeNumber > 4000 else max(0, self.epsilon - ((1.0) / 4000))

if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument("--id", type=int, default=0)
    parser.add_argument("--numOpponents", type=int, default=0)
    parser.add_argument("--numTeammates", type=int, default=0)
    parser.add_argument("--numEpisodes", type=int, default=500)

    args = parser.parse_args()

    # Init Connections to HFO Server
    # hfoEnv = HFOAttackingPlayer(
    #     numOpponents=args.numOpponents, numTeammates=args.numTeammates, agentId=args.id
    # )
    # hfoEnv.connectToServer()
    hfoEnv = hfoEnv()
    log_freq = 50
    rewards_list = []
    rewards = 0
    # init MonteCarloAgent
    agent = MonteCarloAgent(discountFactor=0.99, epsilon=0.6)
    numEpisodes = args.numEpisodes
    numTakenActions = 0
    print(numEpisodes)

    for episode in range(numEpisodes):
        agent.reset()
        observation = hfoEnv.reset()
        status = 0

        # forward, one episode
        while status == 0:  # loop and record one episode,  status = 1 GOAL or OUT
            epsilon = agent.computeHyperparameters(numTakenActions, episode)
            agent.setEpsilon(epsilon)
            agent.setState(agent.toStateRepresentation(observation))
            action = agent.act()
            numTakenActions += 1
            nextObservation, reward, done, status = hfoEnv.step(action, observation)
            agent.setExperience(
                agent.toStateRepresentation(observation),
                action,
                reward,
                status,
                agent.toStateRepresentation(nextObservation)
            )
            observation = nextObservation
            rewards += reward

        # learn, backward
        agent.learn()  # learn from one episode
        if (episode + 1) % log_freq == 0:
            print (episode + 1)
            print (rewards)
            print ("rewards %f" % (rewards/log_freq))
            rewards_list.append(rewards/log_freq)
            rewards = 0

        # draw
    import matplotlib.pyplot as plt
    print(len(list(range(0,numEpisodes,log_freq))))
    print(len(rewards_list))
    plt.plot(list(range(0,numEpisodes,log_freq)), rewards_list)
    plt.savefig("MonteCarlo.png")
    plt.show()